<?php

namespace Drupal\module_builder\Form;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Element;
use DrupalCodeBuilder\Exception\InvalidInputException;

/**
 * Form showing generated component code.
 */
class ComponentGenerateForm extends ComponentFormBase {

  use MessengerTrait;

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $module = $this->entity;
    $component_data = $module->get('data');
    //dsm($component_data);

    // Add in the component root name and readable name.
    $component_data['root_name'] = $module->id;
    $component_data['readable_name'] = $module->label;

    // Build list.
    // The UI always gets the full code.
    $component_data['requested_build'] = array('all' => TRUE);

    // Get the files.
    try {
      $files = $this->codeBuilderTaskHandlerGenerate->generateComponent($component_data);
    }
    catch (InvalidInputException $e) {
      $this->messenger()->addError(t("Invalid input for code generator: @message", [
        '@message' => $e->getMessage(),
      ]));

      return $form;
    }

    // Get the path to the module if it's previously been written.
    $existing_module_path = $this->getExistingModule();
    if ($existing_module_path) {
      $this->messenger()->addWarning(t("This module already exists at @path. It is recommended you use version control to prevent losing any existing code.", [
        '@path' => $existing_module_path,
      ]));
    }

    $module_name = $this->entity->id();
    if (\Drupal::moduleHandler()->moduleExists($module_name)) {
      $this->messenger()->addWarning(t("This module is currently ENABLED on this site. Writing files MAY CAUSE YOUR SITE TO CRASH."));
    }

    $form['code'] = array(
      '#type' => 'vertical_tabs',
    );

    // Why can't these be nested under the vertical_tabs element?
    $form['files']['#tree'] = TRUE;

    ksort($files);
    foreach ($files as $filename => $code) {
      $form['files'][$filename] = array(
        '#type' => 'details',
        '#title' => $filename,
        '#group' => 'code',
      );

      $title = t('@filename code', [
        '@filename' => $filename,
      ]);

      // Warn the user if the file already exists, and is not committed to git.
      $filepath = $existing_module_path . '/' . $filename;
      if (file_exists($filepath)) {
        // TODO: if the file is clean in git, we can maybe skip this, or say
        // it's ok to overwrite?
        $title .= ' ' . t("(File already exists)");

        $git_status = shell_exec("git status {$filepath} --porcelain");

        if (!empty($git_status)) {
          $git_status_code = substr($git_status, 0, 2);

          // TODO: These don't take into account that changes might be staged.
          switch ($git_status_code) {
            case '??':
              $title .= ' ' . t("WARNING: file is not under git version control! Writing this will lose the existing version.");
              break;

            case ' M':
              $title .= ' ' . t("WARNING: file has uncommitted changes! Writing this will lose these.");
              break;
          }
        }
      }

      $form['files'][$filename]['module_code'] = array(
        '#type' => 'textarea',
        '#title' => $title,
        '#rows' => count(explode("\n", $code)),
        '#default_value' => $code,
        // This creates an item in form values that just contains the code, so
        // we don't need to filter out the values from button labels when
        // writing code.
        '#parents' => ['file_code', $filename],
        '#prefix' => '<div class="module-code">',
        '#suffix' => '</div>',
      );

      // We don't actually use the value from the form, as the POST process
      // seems to be turning unix line endings into Windows line endings! Store
      // it in the form state instead.
      $form_state->set(['files', $filename], $code);

      $form['files'][$filename]['write'] = array(
        '#type' => 'submit',
        // WTF: FormBuilder::buttonWasClicked() won't work with dots in the
        // button's #name!
        '#name' => 'write_single_' . str_replace('.', '_', $filename),
        '#value' => $this->t("Write file $filename"),
        '#submit' => array('::submitWriteSingle'),
      );
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = [];

    if (isset($form['files']) && count(Element::children($form['files']))) {
      $actions['write'] = array(
        '#type' => 'submit',
        '#name' => 'write_all',
        '#value' => $this->t('Write all files'),
        '#submit' => array('::write'),
      );
    }

    return $actions;
  }

  /**
   * Returns the path to the module if it has previously been written.
   *
   * @return
   *  A Drupal-relative path to the module folder, or NULL if the module
   *  does not already exist.
   */
  protected function getExistingModule() {
    $module_name = $this->entity->id();

    $exists = \Drupal::service('extension.list.module')->exists($module_name);
    if ($exists) {
      $module = \Drupal::service('extension.list.module')->get($module_name);

      return $module->getPath();
    }
  }

  /**
   * Submit callback to write the module files.
   */
  public function write(array $form, FormStateInterface $form_state) {
    $values = $form_state->getValues();

    $drupal_relative_module_dir = $this->getRelativeModuleFolder();

    \Drupal::service('file_system')->prepareDirectory($path, FileSystemInterface::CREATE_DIRECTORY);

    $count_written = 0;
    foreach (array_keys($values['file_code']) as $module_relative_filepath) {
      $file_contents = $form_state->get(['files', $module_relative_filepath]);

      $result = $this->writeSingleFile($drupal_relative_module_dir, $module_relative_filepath, $file_contents);

      if ($result) {
        $count_written++;
      }
      else {
        $this->messenger()->addError(t("Problem writing file @file", [
          '@file' => $module_relative_filepath
        ]));
      }
    }

    if ($count_written) {
      $this->messenger()->addStatus(t("Written @count files to folder @folder.", [
        '@count'  => $count_written,
        '@folder' => $drupal_relative_module_dir,
      ]));
    }
  }

  /**
   * Submit handler to write a single file.
   */
  public function submitWriteSingle(array $form, FormStateInterface $form_state) {
    $button = $form_state->getTriggeringElement();
    $button_array_parents = $button['#array_parents'];

    $file_key = $button_array_parents[1];

    $file_contents = $form_state->get(['files', $file_key]);

    $drupal_relative_module_dir = $this->getRelativeModuleFolder();

    $result = $this->writeSingleFile($drupal_relative_module_dir, $file_key, $file_contents);

    if ($result) {
      $this->messenger()->addStatus(t("Written file @file to folder @folder.", [
        '@file'  => $file_key,
        '@folder' => $drupal_relative_module_dir,
      ]));
    }
  }

  /**
   * Get the Drupal-relative path of the module folder to write to.
   *
   * @return string
   *   The Drupal-relative path.
   */
  protected function getRelativeModuleFolder() {
    // If the module folder already exists, write there.
    $existing_module_folder = $this->getExistingModule();
    if ($existing_module_folder) {
      return $existing_module_folder;
    }

    $module_name = $this->entity->id();

    if (file_exists('modules/custom')) {
      $modules_dir = 'modules/custom';
    }
    else {
      $modules_dir = 'modules';
    }

    $drupal_relative_module_dir = $modules_dir . '/' . $module_name;

    return $drupal_relative_module_dir;
  }

  /**
   * Writes a single file.
   *
   * @param string $drupal_relative_module_dir
   *   The module folder to write to, as a path relative to Drupal root.
   * @param string $module_relative_filepath
   *   The name of the file to write, as a path relative to the module folder,
   *   e.g. src/Plugins/Block/Foo.php.
   * @param string $file_contents
   *   The file contents to write.
   *
   * @return bool
   *   TRUE if writing succeeded, FALSE if it failed.
   */
  protected function writeSingleFile($drupal_relative_module_dir, $module_relative_filepath, $file_contents) {
    // The files are keyed by a filepath relative to the future module folder,
    // e.g. src/Plugins/Block/Foo.php.
    // Extract the directory.
    $module_relative_dir = dirname($module_relative_filepath);
    $filename = basename($module_relative_filepath);

    $drupal_relative_dir      = $drupal_relative_module_dir . '/' . $module_relative_dir;
    $drupal_relative_filepath = $drupal_relative_module_dir . '/' . $module_relative_filepath;
    \Drupal::service('file_system')->prepareDirectory($drupal_relative_dir, FileSystemInterface::CREATE_DIRECTORY);

    $result = file_put_contents($drupal_relative_filepath, $file_contents);

    return (bool) $result;
  }

}
