<?php

namespace Drupal\Tests\module_builder\Functional;

use Drupal\Tests\BrowserTestBase;
use Drupal\user\RoleInterface;

/**
 * Tests the component edit form.
 *
 * @see \Drupal\module_builder_test_component_type\TestGenerateTask
 *
 * @group module_builder
 */
class ComponentFormTest extends BrowserTestBase {

  /**
   * Disable strict config schema checking.
   *
   * @var bool
   */
  protected $strictConfigSchema = FALSE;

  /**
   * The modules to enable.
   *
   * @var array
   */
  public static $modules = [
    'system',
    'user',
    'block',
    'module_builder',
    'module_builder_test_component_type',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->drupalPlaceBlock('local_actions_block');
    $this->drupalPlaceBlock('local_tasks_block');

    $this->entityTypeManager = $this->container->get('entity_type.manager');
    $this->moduleEntityStorage = $this->entityTypeManager->getStorage('module_builder_module');

    user_role_change_permissions(RoleInterface::ANONYMOUS_ID, [
      'create modules' => TRUE,
    ]);

    $module_entity = $this->moduleEntityStorage->create([
      'id' => 'my_component',
      'name' => 'My Component',
    ]);
    $module_entity->save();
  }

  /**
   * Tests the handling of property default values.
   */
  public function testPropertyDefaults() {
    $web_assert = $this->assertSession();

    // Get to the misc tab where our stuff is.
    $this->drupalGet('/admin/config/development/module_builder/manage/my_component/misc');

    // Check all the elements have their default values.
    $web_assert->fieldValueEquals('data[string_empty]', '');
    $web_assert->fieldValueEquals('data[string_default]', 'default value');
    $web_assert->fieldValueEquals('data[checkbox_empty]', FALSE);
    $web_assert->fieldValueEquals('data[checkbox_default]', TRUE);
    $web_assert->fieldValueEquals('data[array_empty]', '');
    $web_assert->fieldValueEquals('data[array_default]', "value 1\nvalue 2");

    // Submit the form with no changes: defaults should remain the same.
    $this->drupalPostForm(NULL, [], 'Save');

    $web_assert->fieldValueEquals('data[string_empty]', '');
    $web_assert->fieldValueEquals('data[string_default]', 'default value');
    $web_assert->fieldValueEquals('data[checkbox_empty]', FALSE);
    $web_assert->fieldValueEquals('data[checkbox_default]', TRUE);
    $web_assert->fieldValueEquals('data[array_empty]', '');
    $web_assert->fieldValueEquals('data[array_default]', "value 1\nvalue 2");

    $this->moduleEntityStorage->resetCache();
    $module = $this->moduleEntityStorage->load('my_component');

    // The module has the values from the defaults.
    $this->assertEquals('default value', $module->data['string_default']);
    $this->assertEquals(TRUE, $module->data['checkbox_default']);
    $this->assertEquals([
      0 => "value 1",
      1 => "value 2",
    ], $module->data['array_default']);
    // Empty properties stayed empty.
    $this->assertEquals('', $module->data['string_empty']);
    $this->assertEquals(FALSE, $module->data['checkbox_empty']);
    $this->assertEquals([], $module->data['array_empty']);

    // Submit the form with empty values for everything: defaults should be
    // zapped.
    $this->submitForm([
      'data[string_empty]' => '',
      'data[string_default]' => '',
      'data[checkbox_empty]' => FALSE,
      'data[checkbox_default]' => FALSE,
      'data[array_empty]' => '',
      'data[array_default]' => '',
    ], 'Save');

    $this->moduleEntityStorage->resetCache();
    $module = $this->moduleEntityStorage->load('my_component');

    // The module's values are now all empty.
    $this->assertEquals('', $module->data['string_empty']);
    $this->assertEquals('', $module->data['string_default']);
    $this->assertEquals(FALSE, $module->data['checkbox_empty']);
    $this->assertEquals(FALSE, $module->data['checkbox_default']);
    $this->assertEquals([], $module->data['array_empty']);
    $this->assertEquals([], $module->data['array_default']);

    $web_assert->fieldValueEquals('data[string_empty]', '');
    $web_assert->fieldValueEquals('data[string_default]', '');
    $web_assert->fieldValueEquals('data[checkbox_empty]', FALSE);
    $web_assert->fieldValueEquals('data[checkbox_default]', FALSE);
    $web_assert->fieldValueEquals('data[array_empty]', '');
    $web_assert->fieldValueEquals('data[array_default]', "");
  }

  /**
   * Tests the handling of compound properties.
   */
  public function testCompoundProperties() {
    $web_assert = $this->assertSession();
    $page = $this->getSession()->getPage();

    $this->drupalGet('/admin/config/development/module_builder/manage/my_component/misc');

    $page->pressButton('Add a compound empty item');
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][0][one]']);
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][0][two]']);

    $page->pressButton('Add another compound empty item');
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][1][one]']);
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][1][two]']);

    $page->pressButton('Remove last item');
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][0][one]']);
    $web_assert->elementExists('named_exact', ['id_or_name', 'data[compound_empty][0][two]']);
    $web_assert->elementNotExists('named_exact', ['id_or_name', 'data[compound_empty][1][one]']);
    $web_assert->elementNotExists('named_exact', ['id_or_name', 'data[compound_empty][1][two]']);

    // Save the module with one compound item.
    $this->submitForm([
      'data[compound_empty][0][one]' => 'value one',
      'data[compound_empty][0][two]' => 'value two',
    ], 'Save');

    $this->moduleEntityStorage->resetCache();
    $module = $this->moduleEntityStorage->load('my_component');

    $this->assertEquals('value one', $module->data['compound_empty'][0]['one']);
    $this->assertEquals('value two', $module->data['compound_empty'][0]['two']);
    $this->assertArrayNotHasKey(1, $module->data['compound_empty']);

    // Remove the item we saved, and save again: it should be removed from the
    // module.
    $page->pressButton('Remove last item');
    $this->submitForm([], 'Save');

    $this->moduleEntityStorage->resetCache();
    $module = $this->moduleEntityStorage->load('my_component');

    $this->assertArrayNotHasKey(0, $module->data['compound_empty']);
  }

}
